from __future__ import annotations

# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.

EXT_M3U = "#EXTM3U"
EXT_X_TARGET_DURATION = "#EXT-X-TARGETDURATION"
EXT_X_MEDIA_SEQUENCE = "#EXT-X-MEDIA-SEQUENCE"
EXT_X_DISCONTINUITY_SEQUENCE = "#EXT-X-DISCONTINUITY-SEQUENCE"
EXT_X_PROGRAM_DATE_TIME = "#EXT-X-PROGRAM-DATE-TIME"
EXT_X_MEDIA = "#EXT-X-MEDIA"
EXT_X_PLAYLIST_TYPE = "#EXT-X-PLAYLIST-TYPE"
EXT_X_KEY = "#EXT-X-KEY"
EXT_X_STREAM_INF = "#EXT-X-STREAM-INF"
EXT_X_VERSION = "#EXT-X-VERSION"
EXT_X_ALLOW_CACHE = "#EXT-X-ALLOW-CACHE"
EXT_X_END_LIST = "#EXT-X-ENDLIST"
EXT_INF = "#EXTINF"
EXT_I_FRAMES_ONLY = "#EXT-X-I-FRAMES-ONLY"
EXT_X_BIT_RATE = "#EXT-X-BITRATE"
EXT_X_BYTE_RANGE = "#EXT-X-BYTERANGE"
EXT_X_I_FRAME_STREAM_INF = "#EXT-X-I-FRAME-STREAM-INF"
EXT_X_DISCONTINUITY = "#EXT-X-DISCONTINUITY"
EXT_X_CUE_OUT = "#EXT-X-CUE-OUT"
EXT_X_CUE_OUT_CONT = "#EXT-X-CUE-OUT-CONT"
EXT_X_CUE_IN = "#EXT-X-CUE-IN"
EXT_X_CUE_SPAN = "#EXT-X-CUE-SPAN"
EXT_X_SCTE35 = "#EXT-OATCLS-SCTE35"
EXT_IS_INDEPENDENT_SEGMENTS = "#EXT-X-INDEPENDENT-SEGMENTS"
EXT_X_MAP = "#EXT-X-MAP"
EXT_X_START = "#EXT-X-START"
EXT_X_SERVER_CONTROL = "#EXT-X-SERVER-CONTROL"
EXT_X_PART_INF = "#EXT-X-PART-INF"
EXT_X_PART = "#EXT-X-PART"
EXT_X_RENDITION_REPORT = "#EXT-X-RENDITION-REPORT"
EXT_X_SKIP = "#EXT-X-SKIP"
EXT_X_SESSION_DATA = "#EXT-X-SESSION-DATA"
EXT_X_SESSION_KEY = "#EXT-X-SESSION-KEY"
EXT_X_PRELOAD_HINT = "#EXT-X-PRELOAD-HINT"
EXT_X_DATERANGE = "#EXT-X-DATERANGE"
EXT_X_GAP = "#EXT-X-GAP"
EXT_X_CONTENT_STEERING = "#EXT-X-CONTENT-STEERING"
